xquery version "3.1";

module namespace search-facets="http://grammateus.ch/search-facets";

import module namespace templates="http://exist-db.org/xquery/templates";
import module namespace app="http://grammateus.ch/templates" at "app.xql";
import module namespace config="http://grammateus.ch/config" at "config.xqm";
import module namespace gramm="http://grammateus.ch/gramm" at "gramm.xql";
import module namespace show="http://grammateus.ch/show" at "show.xql";
import module namespace URI="http://grammateus.ch/URI" at "uri.xql";

import module namespace console="http://exist-db.org/xquery/console";

declare namespace tei = "http://www.tei-c.org/ns/1.0";

(: Remark: the keyword and date filters are not facets. :)
declare variable $search-facets:list :=
    <list>
        <item id="f-type">Type</item>
        <item id="f-subtype">Subtype</item>
        <item id="f-variation">Variation</item>
        <item id="shape">Shape of Papyrus</item>
        <item id="fibres">Direction of Fibres</item>
        <item id="origPlace">Provenance</item>
    </list>;
 
(: sorting a map - from vangogh letters. 
 : https://gitlab.existsolutions.com/tei-publisher/vangogh/blob/master/modules/facets.xql#L24-35 :)   
declare function local:sort($facets as map(*)?) {
    array {
        if (exists($facets)) then
            for $key in map:keys($facets)
            let $value := map:get($facets, $key)
            order by $key ascending
            return
                map { $key: $value }
        else
            ()
    }
};

(:~ calculates the range of 5 page numbers to be displayed for pagination.
 : 
 : @param page - current page nb
 : @param pagelength - total nb of pages
 :)
declare function local:page-range($page as xs:integer, $page_total as xs:integer) {
    
    let $range :=
        if ($page_total <= 5) then (1 to $page_total)
        else if ($page < 4 and $page + 2 <= $page_total ) then (1 to 5)
        else if (($page + 2) > $page_total) then ($page_total - 4 to $page_total)
        else ($page - 2 to $page + 2)
    
    return $range
};

(:~ returns pagination links.
 : 
 : @param results - query results
 : @param pagelength - number of results per page
 :)
declare function local:pagination($results, $pagelength as xs:integer) {
    
    let $page_current := number(request:get-parameter("page", "1"))
    
    let $page_total := (count($results) idiv $pagelength)+1
    
    let $query := request:get-query-string() => replace("page=.+?&amp;", "")
    
    let $pagination := if (count($results) <= $pagelength) then () 
    else
    <div>
        <nav aria-label="Page navigation">
            <ul class="pagination papyri-pagination">
                <li class="page-item papyri-page-item {if($page_current = 1) then "disabled" else ()}">
                    <a class="page-link papyri-page-link" href="?{$query}page={$page_current - 1}&amp;" aria-label="Previous">
                        <span aria-hidden="true">{gramm:fa-previous()}</span>
                        <span class="sr-only">Previous</span>
                    </a>
                </li>
                
                { for $i in local:page-range($page_current, $page_total)
                    return
                        <li  class="page-item papyri-page-item {if ($i = $page_current) then "active" else ()}">
                            <a class="page-link papyri-page-link" href="?{$query}page={$i}&amp;">{$i}</a>
                        </li>
                }
                   
                <li class="page-item papyri-page-item {if($page_current = $page_total) then "disabled" else ()}">
                    <a class="page-link papyri-page-link" href="?{$query}page={$page_current + 1}&amp;" aria-label="Next">
                        <span aria-hidden="true">{gramm:fa-next()}</span>
                        <span class="sr-only">Next</span>
                    </a>
                </li>
            </ul>
        </nav>
    </div>
    
    return $pagination
};

(:~ returns message with number of results for the query.
 : 
 : @param results - query results
 : @param page - current result page
 : @param pagelength - number of results per page
 :)
declare function local:results-nb($results, $pagelength as xs:integer, $page as xs:integer) {
    let $nb := count($results)
    let $message := 
        if ($nb <= $pagelength) 
        then let $pap := if ($nb > 1) then " Papyri" else " Papyrus"
        return $nb || $pap
        else 
            let $start := ($pagelength*($page -1))+1
            let $end := $start -1 + count(subsequence($results, ($pagelength*($page -1))+1, $pagelength))
            return $start || "-" || $end || " of " || $nb || " Papyri"
    
    return $message
};

(:~ returns formated search result for the papyri page.
 : 
 : @param r - an XML EpiDoc document from the database
 :)
declare function search-facets:format-results($r) {
    <div class="result row border-bottom">
        <div class="col result-tm"><!-- col-md-2 -->
            <span class="badge badge-secondary">TM {gramm:tm($r)}</span>
        </div>
        <div class="col result-info"><!-- col-md-10 -->
            <!-- Link -->
            {(gramm:get-doc-link($r), if (gramm:get-iiif-manifest($r)) then gramm:iiif-logo() else ())}
            
            <!-- Typology -->
            <div>Type: {gramm:get-catDesc($r//tei:catRef[@n eq 'format-type'])}</div>
            <div>Subtype: {gramm:get-catDesc($r//tei:catRef[@n eq 'format-subtype'])}</div>
            {
                let $variation := gramm:get-catDesc($r//tei:catRef[@n eq 'format-variation'])
                return
                    if ($variation) 
                    then <div>Variation: {$variation}</div> 
                    else()
            }
            
            <!-- Date/Place -->
            <div>
                <!--<span>{gramm:fa-calendar()}&#160;</span>-->
                <span>{show:date($r)},&#160;</span>
                <!--<span>{gramm:fa-map()}&#160;</span>-->
                <span>{gramm:origPlace($r)}</span>
            </div>
            
            <!-- Shape/Fibres -->
            <span>Shape&#160;&#160;</span><span>{
                ($r//tei:width div $r//tei:height) => gramm:shape() => gramm:shape-symbol()
            }&#160;&#160;</span>

            <span>Fibres {
                gramm:fibres($r) => gramm:fibres-symbol()
            }</span>
            
        </div>
    </div>
};

(:~ returns the html (bootstrap card) for each dimension in the variable $search-facet:list
 : 
 : @param results - a sequence of XML documents
 :)
declare function search-facets:get-facets($results){
    for $dimension in $search-facets:list/item
    let $facets := ft:facets($results, string($dimension/@id), ())
    return
        <div class="card">
            <div class="card-header facet-header" id="heading-{$dimension/@id}">
                <button class="btn btn-link mr-auto" type="button" data-toggle="collapse" data-target="#{$dimension/@id}" aria-expanded="true" aria-controls="{$dimension/@id}">
                    <span class="mr-auto">{$dimension/string()}</span>
                        {(:first facet card is open :)
                        (:if ($dimension/@id eq $search-facets:list/item[1]/@id) then gramm:fa-up() else:) gramm:fa-down()}
                </button>
            </div>
            <!--first facet open:
            <div id="{$dimension/@id}" class="collapse {if ($dimension/@id eq $search-facets:list/item[1]/@id) then 'show' else ()}" aria-labelledby="heading-{$dimension/@id}">-->
            <div id="{$dimension/@id}" class="collapse" aria-labelledby="heading-{$dimension/@id}">
                <div class="card-body">
                {array:for-each(local:sort($facets), function($entry) {
                    map:for-each($entry, function($label, $count) {
                        <div class="facet-content">
                            <input class="form-check-input" type="checkbox" name="facet:{$dimension/@id}" value="{$label}"/>
                            <label>{$label} <span style="font-style:italic;">({$count})</span></label>
                        </div>
                        })
                    })
                }
                </div>
            </div>
    </div>
};


(: ---------- TEMPLATE FUNCTION (search entry point) ---------- :)

declare
%templates:default("keyword", "")
%templates:default("pagelength", 100)
function search-facets:advanced-search ($node as node (), $model as map (*), $keyword as xs:string, $pagelength as xs:integer){
    
    (: current page :)
    let $page := number(request:get-parameter("page", "1"))
    
    (: get list of all existing parameters:)
    let $params := request:get-parameter-names()
    
    (: create the facet options from existing parameters :)
    let $options := map{
        "facets" : map:merge(
            (:add only facet parameters to the map:)
            for $p in $params
            where matches($p, "facet:")
            return map { substring-after($p, "facet:"): request:get-parameter($p, "")}
        )
    }
    
    let $results_kw := 
        (: filter docs by keyword :)
        if ($keyword != '') then collection($URI:papyri)//tei:TEI[ft:query(., $keyword, $options)]
        (: or else search all documents :)
        else collection($URI:papyri)//tei:TEI[ft:query(., (), $options)]
    
    (: filter papyri collection for the date :)
    let $results_date :=
        if (["min-date", "max-date"] = $params) 
        then 
            let $min := if ("min-date" = $params) then request:get-parameter("min-date", ()) else "-300"
            let $max := if ("max-date" = $params) then request:get-parameter("max-date", ()) else "300"
            return gramm:date-filter($min, $max, $results_kw)
        else $results_kw
    
    (: filter results for the keyword :)
(:    let $results_kw := :)
(:        (: filter docs by keyword :):)
(:        if ($keyword != '') then $results_date[ft:query(., $keyword, $options)]:)
(:        (: or else search all documents :):)
(:        else $results_date[ft:query(., (), $options)]:)
        
    (: sort result :)
    let $order-by := request:get-parameter("sort-by", "")
    let $sort := 
            if ($order-by eq 'type') then 'for $r in $results_date
            order by gramm:get-catDesc($r//tei:catRef[@n eq "format-type"]),
            gramm:get-catDesc($r//tei:catRef[@n eq "format-subtype"]),
            gramm:get-catDesc($r//tei:catRef[@n eq "format-variation"])
            return $r'
            else if ($order-by eq 'place') then 'for $r in $results_date 
            order by gramm:origPlace($r//tei:history) return $r'
            else if ($order-by eq 'TM') then 'for $r in $results_date order by gramm:tm($r) => number() return $r'
            else 'for $r in $results_date
            let $vol := if ($r//tei:biblScope[@unit eq "volume"]) then number($r//tei:biblScope[@unit eq "volume"]) else number("0")
            order by $r//tei:title[@level eq "s"], $vol, number($r//tei:biblScope[@unit eq "numbers"]) ascending, $r//tei:biblScope[@unit eq "generic"] 
            return $r'

    let $results := util:eval($sort)
    
    (: get labels/count for each facet dimension :)
    let $facets := search-facets:get-facets($results)

    return 
    <div class="col search">
        <div class="row">
            <div class="col">
                <ul class="legend">
                    {   (: button to remove all paramters from query string :)
                        if (count($params) > 0) 
                        then 
                            <li>
                                <a class="badge badge-danger remove-filters" href="papyri.html">
                                    Remove all filters {gramm:fa-times()}
                                </a>
                            </li>
                        else ()}
                    {   (: parameters to remove from current query :)
                        let $current := request:get-query-string() => replace("%20", " ")
                        for $p in $params
                            (: parameter and value as a string :)
                            let $value := request:get-parameter($p, ())
                            (: there can be several values for the same parameter name :)
                            for $v in $value
                                (: last & is to mark the next param, so that we do not replace a similar one
                                    => esp. for origPlace
                                    => for that we also add a final &, see papyri.html l.131:)
                                let $string := $p || "=" || $v || "&amp;"
                                let $regex_1 := replace($string, "\(", "\\(")
                                let $regex_2 := replace($regex_1, "\)", "\\)")
                                let $regex := replace($regex_2, "\?", "%3F")
                                (: new query without the parameter :)
                                let $new_query_1 := replace($current, $regex, "")
                                (: we remove the page parameter from the new query :)
                                let $new_query := replace($new_query_1, "page=.+?&amp;", "")
                                (: parameter name for display :)
                                let $param_name := 
                                    if (matches($p, 'facet:'))
                                    (: facet parameters :)
                                    then $search-facets:list/item[@id eq substring-after($p, "facet:")]/string()
                                    (: "keyword" parameter :)
                                    else $p
                                return
                                    (: ignore the page parameter :)
                                    if ($p = "page") then () else
                                    <li>
                                        <a class="badge badge-secondary" href="papyri.html?{$new_query}">
                                            {$param_name || ": " || $v || " "}{gramm:fa-times()}
                                        </a>
                                    </li>
                    }
                </ul>
            </div>
        </div>
        <div class="row">
            <div class="col-md-4 papyri-col">
                <h3 class="h3">Filters</h3>
                <form action="papyri.html?" method="get" class="facetForm">
                    <!-- keyword search -->
                    <div class="card">
                        <div class="card-header facet-header" id="heading-keyword">
                            <button class="btn btn-link mr-auto" type="button" data-toggle="collapse" data-target="#keyword" aria-expanded="true" aria-controls="keyword">
                                <span class="mr-auto">Keyword</span>
                                {gramm:fa-up()}
                            </button>
                        </div>
                        <div id="keyword" class="collapse show" aria-labelledby="heading-keyword">
                            <div class="card-body" style="min-height:100px;">
                                <input id="doc-kw" name="keyword" type="text" class="form-control"
                               placeholder="e.g. 'Camel', or a TM number" value="{$keyword}" aria-label="Search"/>
                               <p>{gramm:fa-info()} Search for HGV subjects or any keyword.</p>
                            </div>
                        </div>
                    </div>
                    <!-- facets from index -->
                    {$facets}
                    <!-- date range slider -->
                    <div class="card">
                        <div class="card-header facet-header" id="heading-century">
                            <button class="btn btn-link mr-auto" type="button" data-toggle="collapse" data-target="#century" aria-expanded="true" aria-controls="century">
                                <span class="mr-auto">Century</span>
                                {gramm:fa-down()}
                            </button>
                        </div>
                        <div id="century" class="collapse" aria-labelledby="heading-century">
                            <div class="card-body" style="min-height:100px;">
                                <div id="date-range" class="noUiSlider"></div>
                                <input type="hidden" id="min-date" name="min-date"/>
                                <input type="hidden" id="max-date" name="max-date"/>
                            </div>
                        </div>
                    </div>
                    <button class="btn btn-light" type="submit">Search</button>
                </form>
            </div>
            <div class="col-md-8 papyri-col">
                <h3 class="h3 row">
                    <div class="col">
                        {
                            local:results-nb($results, $pagelength, $page)
                        } 
                    </div>
                    <div class="col d-flex justify-content-end flex-wrap">
                        <div class="dropdown order-by">
                            <a class="dropdown-toggle sort-papyri" data-toggle="dropdown" href="#">Sort by
                                <span class="caret"></span>
                            </a>
                            <div class="dropdown-menu" aria-labelledby="dropdownMenuButton">
                            {
                                let $params := request:get-query-string() => replace("sort-by=.+?&amp;", "")
                                return 
                                <div>
                                    <a class="dropdown-item" href="{concat('papyri.html?',$params,"sort-by=publication&amp;")}">Publication</a>
                                    <a class="dropdown-item" href="{concat('papyri.html?',$params,'sort-by=type&amp;')}">Type</a>
                                    <a class="dropdown-item" href="{concat('papyri.html?',$params,'sort-by=place&amp;')}">Place</a>
                                    <a class="dropdown-item" href="{concat('papyri.html?',$params,'sort-by=TM&amp;')}">TM</a>
                                </div>
                            }
                            </div>
                        </div>
                        <!-- pagination -->
                        {local:pagination($results, $pagelength)}
                    </div>
                </h3>
                <div id="search-results">
                    {
                        (: subsequence: results, starting position, number of items:)
                        for $r in subsequence($results, ($pagelength*($page -1))+1, $pagelength)
                        return
                            search-facets:format-results($r)
                    }
                </div>
                <!-- pagination -->
                <div class="row bottom-pagination d-flex justify-content-end">
                    {local:pagination($results, $pagelength)}
                </div>
                
                {gramm:back-to-top()}
            </div>
        </div>
    </div>
};